//
// decimal.cpp
//
// Bill Seymour, 2024-02-02
//
// Copyright Bill Seymour 2024.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This TU contains definitions of the non-inline functions
// declared in decimal.hpp.
//

#include "decimal.hpp"

#include <algorithm> // reverse
#include <stdexcept>

#include <cstdlib> // div, div_t
#include <cstring> // strcmp
#include <cctype>  // isdigit
#include <cfenv>
#include <cassert>

namespace
{
    using std::size_t;
    using std::move;
    using std::string;
    using std::numeric_limits;
    using std::div;
    using std::div_t;

    //
    // For converting to/from integers:
    //
    const bignum::integer bigten(10U);
}

namespace bignum {

void decimal::init(unsigned long long value)
{
    assert(val.empty());

    while (value != 0)
    {
        val.push_back(value % 10U);
        value /= 10U;
    }
    trim_trailing_zeros();
}

void decimal::init(long long value)
{
    if (value < 0)
    {
        neg = true;
        if (value != LLONG_MIN || LLONG_MIN == -LLONG_MAX)
        {
            value = -value;
        }
    }
    init(static_cast<unsigned long long>(value));
}

void decimal::init(long double value)
{
    assert(val.empty());

    if (!std::isfinite(value))
    {
        throw std::invalid_argument("Can't convert from non-finite values.");
    }
    if (value < 0.0L)
    {
        neg = true;
        value = -value;
    }

    //
    // First, make it an integer adjusting this_scale as appropriate.
    //
    while (std::fmod(value, 1.0L) > 0.0L && this_scale <= max_digs)
    {
        value *= 10.0L;
        ++this_scale;
    }

    //
    // Now we can save the value.
    //
    while (value >= 1.0L)
    {
        long double rem = std::floor(std::fmod(value, 10.0L));
        val.push_back(static_cast<elem_type>(rem));
        value = std::floor(value / 10.0L);
    }
    trim_trailing_zeros();
}

void decimal::init(const char* s, const char** term)
{
    assert(val.empty());

    if (nullptr == s)
    {
        throw std::invalid_argument("Can't convert to decimal from nullptr");
    }

    user_scale = this_scale = 0;
    neg = false;

    if (*s == '-')
    {
        neg = true;
        ++s;
    }
    else if (*s == '+')
    {
        ++s;
    }

    for (bool got_point = false; *s != '\0'; ++s)
    {
        if (*s == '.')
        {
            if (got_point)
            {
                break;
            }
            got_point = true;
            continue;
        }
        if (*s == 'e' || *s == 'E')
        {
            this_scale -= atoi(++s);
            if (term != nullptr)
            {
                if (*s == '+' || *s == '-')
                {
                    ++s;
                }
                while (std::isdigit(static_cast<unsigned char>(*s)))
                {
                    ++s;
                }
            }
            break;
        }
        if (!std::isdigit(static_cast<unsigned char>(*s)))
        {
            break;
        }
        if (got_point)
        {
            ++this_scale;
        }
        val.push_front(*s - '0');
    }
    if (term != nullptr)
    {
        *term = s;
    }
    trim_all_zeros();
    user_scale = this_scale;
}

decimal::decimal(const integer& value, scale_type s, rounding r)
  : user_scale(s), rnd(r)
{
    integer temp(value);
    if (temp.is_neg())
    {
        this->neg = true;
        temp.negate();
    }

    while (!temp.is_zero())
    {
        integer::div_t d = temp.div(bigten);
        val.push_back(static_cast<elem_type>(d.rem));
        temp = move(d.quot);
    }
    trim_trailing_zeros();
}

decimal decimal::rawval() const
{
    decimal temp(*this);
    temp.this_scale = 0;
    return temp;
}

int decimal::signum() const noexcept
{
    return iszero() ? 0 : isneg() ? -1 : +1;
}

integer decimal::to_integer(rounding r) const
{
    decimal temp(*this);
    temp.user_scale = 0;
    temp.set_rounding(r);
    temp.round();
    if (temp.this_scale > 0)
    {
        temp.val.insert(temp.val.end(), this_scale, 0);
    }

    integer retval;
    for (criter it = temp.val.crbegin(), end = temp.val.crend(); it != end; ++it)
    {
        retval *= bigten;
        retval += *it;
    }
    return retval;
}

std::string decimal::to_string(int prec, bool sci) const
{
    if (iszero())
    {
        return "0";
    }

    decimal temp(*this);
    temp.round();
    int siz = static_cast<int>(temp.val.size());

    if (prec > max_digs || prec <= 0)
    {
        prec = max_digs;
    }
    if (!isfloat() && temp.this_scale > temp.user_scale)
    {
        int tmp = siz - (temp.this_scale - temp.user_scale);
        if (prec > tmp)
        {
            prec = tmp;
        }
    }
    if (siz > prec)
    {
        temp.do_round(prec, default_rounding);
        siz = static_cast<int>(temp.val.size());
    }
    string retval;
    retval.reserve(siz + 5);

    //
    // Force scientific notation if siz is too small to write an integer,
    // i.e., if the number of digits is not greater than the number
    // to the right of the decimal point.
    //
    while (siz < prec && temp.this_scale < 0)
    {
        temp.val.push_front(0);
        ++temp.this_scale;
        ++siz;
    }
    if (sci || temp.this_scale < 0)
    {
        if (temp.neg)
        {
            retval.push_back('-');
        }
        if (1 == siz)
        {
            retval.push_back(static_cast<char>(temp.val.front()) + '0');
        }
        else
        {
            bool need_point = true;
            criter beg = temp.val.crbegin();
            criter const end = temp.val.crend();
            for ( ; beg != end; ++beg)
            {
                retval.push_back(static_cast<char>(*beg) + '0');
                if (need_point)
                {
                    retval.push_back('.');
                    need_point = false;
                }
            }
        }

        retval.push_back('e');

        int exp = siz - 1 - temp.this_scale;
        if (exp < 0)
        {
            retval.push_back('-');
            exp = -exp;
        }
        else
        {
            retval.push_back('+');
        }

        if (exp < 10) // at least 2 digits
        {
            retval.push_back('0');
            retval.push_back(static_cast<char>(exp) + '0');
        }
        else if (exp < 100)
        {
            std::div_t dexp = std::div(exp, 10);
            retval.push_back(static_cast<char>(dexp.quot) + '0');
            retval.push_back(static_cast<char>(dexp.rem) + '0');
        }
        else // we need more than 2 digits
        {
            retval.append(std::to_string(exp));
        }
    }
    else
    {
        bool got_point = false;
        int  out_digs = 0;

        if (temp.this_scale < 0)
        {
            retval.assign(-temp.this_scale, '0');
        }
        for (elem_type elem : temp.val) // little-endian, so digits in reverse
        {
            if (!got_point && out_digs == temp.this_scale && out_digs > 0)
            {
                retval.push_back('.');
                got_point = true;
            }
            retval.push_back(static_cast<char>(elem + '0'));
            ++out_digs;
        }
        if (out_digs < temp.this_scale)
        {
            retval.append(temp.this_scale - out_digs, '0');
            retval.append(".0");
        }
        else if (temp.this_scale > 0 && siz == temp.this_scale)
        {
            retval.append(".0");
        }
        if (temp.neg)
        {
            retval.push_back('-');
        }
        std::reverse(retval.begin(), retval.end());
    }

    return retval;
}

void decimal::inc()
{
    signed char carry = 1;
    for (iter it = val.begin(), end = val.end(); carry != 0 && it != end; ++it)
    {
        *it += carry;
        carry = 0;
        if (*it >= 10)
        {
            *it -= 10;
            carry = 1;
        }
    }
    if (carry != 0)
    {
        val.push_back(1);
    }
}

void decimal::do_round(int prec, rounding r)
{
    if (prec <= 0) // just in case...
    {
        clear();
        return;
    }
    if (prec > max_digs)
    {
        prec = max_digs;
    }

    int siz = static_cast<int>(val.size());

    if (isfloat() || siz <= prec)
    {
        return;
    }

    //
    // Get rid of all but a guard digit:
    //
    bool had_less_sig_digs = false;
    siz -= prec + 1;
    if (siz > 0)
    {
        had_less_sig_digs = true;
        this_scale -= siz;
        do {
            val.pop_front();
        } while (--siz > 0);
    }

    //
    // The guard digit:
    //
    bool lt = false, tie = false, gt = false;
    if (val.empty())
    {
        //
        // There isn't one, but we've had less significant non-zero digits.
        //
        lt = had_less_sig_digs;
    }
    else
    {
        switch(val.front())
        {
            case 9:
            case 8:
            case 7:
            case 6:
                gt = true;
                break;
            case 5:
                if (had_less_sig_digs)
                {
                    gt = true;
                }
                else
                {
                    tie = true;
                }
                break;
            default:
                lt = had_less_sig_digs;
                break;
        }
        val.pop_front();
        --this_scale;
    }

    //
    // Whether the remaining LSD is odd or even:
    //
    bool zer = val.empty();
    if (zer)
    {
        neg = false;
    }
    bool odd = zer ? false : (val.front() & 1) != 0;

    switch(r)
    {
        case rounding::all_to_neg_inf:
            if (neg)
                inc();
            break;
        case rounding::all_to_pos_inf:
            if (!neg)
                inc();
            break;
        case rounding::all_to_zero:
        case rounding::all_fastest:
        case rounding::all_smallest:
        case rounding::all_unspecified:
            break;
        case rounding::all_away_zero:
            inc();
            break;
        case rounding::all_to_odd:
            if (!odd)
                inc();
            break;
        case rounding::all_to_even:
            if (odd)
                inc();
            break;

        case rounding::tie_to_neg_inf:
            if (!neg && gt)
                inc();
            else if (neg && !lt)
                inc();
            break;
        case rounding::tie_to_pos_inf:
            if (!neg && !lt)
                inc();
            else if (neg && gt)
                inc();
            break;
        case rounding::tie_away_zero:
            if (!lt)
                inc();
            break;
        case rounding::tie_to_odd:
            if (gt || tie && !odd)
                inc();
            break;
        case rounding::tie_to_even:
            if (gt || tie && odd)
                inc();
            break;
        default: // tie_fastest, tie_smallest, tie_unspecified, tie_to_zero
            if (gt)
                inc();
            break;
    }

    trim_trailing_zeros();
}

decimal::aligned_digits decimal::align_digits(const decimal& rhs) const
{
    aligned_digits retval { this->val, rhs.val };
    int diff = int(this->this_scale) - int(rhs.this_scale);
    if (diff < 0)
    {
        retval.first.insert(retval.first.begin(), -diff, elem_type(0));
    }
    else if (diff > 0)
    {
        retval.second.insert(retval.second.begin(), +diff, elem_type(0));
    }
    return retval;
}

int decimal::cmp(const absval& lhs, const absval& rhs) noexcept
{
    size_t lsize = lhs.size();
    size_t rsize = rhs.size();
    if (lsize > rsize)
    {
        return +1;
    }
    if (lsize < rsize)
    {
        return -1;
    }
    for (size_t idx = lsize; idx > 0; )
    {
        --idx;
        elem_type diff = lhs[idx] - rhs[idx];
        if (diff > 0)
        {
            return +1;
        }
        if (diff < 0)
        {
            return -1;
        }
    }
    return 0;
}

int decimal::raw_cmp(const absval& lhs, const absval& rhs)
{
    integer lval, rval;
    for (unsigned lidx = lhs.size(); lidx > 0; )
    {
        lval *= bigten;
        lval += lhs[--lidx];
    }
    for (unsigned ridx = rhs.size(); ridx > 0; )
    {
        rval *= bigten;
        rval += rhs[--ridx];
    }
    return lval.compare(rval);
}

int decimal::cmp(const decimal& rhs) const
{
    if (!this->neg && rhs.neg)
    {
        return +1;
    }
    if (this->neg && !rhs.neg)
    {
        return -1;
    }
    aligned_digits digs(move(align_digits(rhs)));
    return cmp(digs.first, digs.second);
}

void decimal::do_inc_dec(bool subtract)
{
    static decimal unity(1U);
    unity.neg = subtract;
    *this += unity;
}

void decimal::add(absval& lhs, const absval& rhs)
{
    std::div_t carry = { 0, 0 };

    iter  lidx = lhs.begin(),  lend = lhs.end();
    citer ridx = rhs.cbegin(), rend = rhs.cend();

    for ( ; lidx != lend && ridx != rend; ++lidx, ++ridx)
    {
        carry = std::div(*lidx + *ridx + carry.quot, 10);
        *lidx = static_cast<signed char>(carry.rem);
    }
    for ( ; carry.quot != 0 && lidx != lend; ++lidx)
    {
        carry = std::div(*lidx + carry.quot, 10);
        *lidx = static_cast<signed char>(carry.rem);
    }
    for ( ; carry.quot != 0 && ridx != rend; ++ridx)
    {
        carry = std::div(*ridx + carry.quot, 10);
        lhs.push_back(static_cast<signed char>(carry.rem));
    }
    for ( ; ridx != rend; ++ridx)
    {
        lhs.push_back(*ridx);
    }
    if (carry.quot != 0)
    {
        lhs.push_back(static_cast<signed char>(carry.quot));
    }
}

void decimal::sub(absval& lhs, const absval& rhs)
{
    elem_type borrow = 0;

    iter  lidx = lhs.begin(),  lend = lhs.end();
    citer ridx = rhs.cbegin(), rend = rhs.cend();

    for ( ; lidx != lend && ridx != rend; ++lidx, ++ridx)
    {
        *lidx -= *ridx + borrow;
        borrow = 0;
        if (*lidx < 0)
        {
            *lidx += 10;
            ++borrow;
        }
    }
    for ( ; borrow != 0 && lidx != lend; ++lidx)
    {
        *lidx -= borrow;
        borrow = 0;
        if (*lidx < 0)
        {
            *lidx += 10;
            ++borrow;
        }
    }
    assert(0 == borrow);
}

decimal& decimal::operator+=(const decimal& rhs)
{
    if (rhs.iszero())
    {
        return *this;
    }
    if (&rhs == this) // 2024-01-25:  guard against aliasing
    {
        mul_part(this->val, 2);
        return *this;
    }
    if (this->iszero())
    {
        this->val = rhs.val;
        this->neg = rhs.neg;
        this->this_scale = rhs.this_scale;
        return *this;
    }

    aligned_digits digs(move(align_digits(rhs)));
    if (rhs.neg == this->neg)
    {
        add(digs.first, digs.second);
        this->val = move(digs.first);
        trim_trailing_zeros();
    }
    else
    {
        if (cmp(digs.first, digs.second) >= 0)
        {
            sub(digs.first, digs.second);
            this->val = move(digs.first);
        }
        else
        {
            sub(digs.second, digs.first);
            this->val = move(digs.second);
            this->neg = !this->neg;
        }
        trim_all_zeros();
    }

    return *this;
}

decimal& decimal::operator-=(const decimal& rhs)
{
    if (rhs.iszero())
    {
        return *this;
    }
    if (&rhs == this) // 2024-01-25:  guard against aliasing
    {
        this->clear();
        return *this;
    }
    if (this->iszero())
    {
        this->val = rhs.val;
        this->neg = !rhs.neg;
        this->this_scale = rhs.this_scale;
        return *this;
    }

    aligned_digits digs(move(align_digits(rhs)));
    if (rhs.neg != this->neg)
    {
        add(digs.first, digs.second);
        this->val = move(digs.first);
        trim_trailing_zeros();
    }
    else
    {
        if (cmp(digs.first, digs.second) >= 0)
        {
            sub(digs.first, digs.second);
            this->val = move(digs.first);
        }
        else
        {
            sub(digs.second, digs.first);
            this->val = move(digs.second);
            this->neg = !this->neg;
        }
        trim_all_zeros();
    }

    return *this;
}

//
// Create a partial product given one multiplier digit.
//
void decimal::mul_part(absval& lhs, elem_type rhs)
{
    int carry = 0;

    for (iter it = lhs.begin(), end = lhs.end(); it != end; ++it)
    {
        std::div_t d = std::div(int(*it) * rhs + carry, 10);
        *it = elem_type(d.rem);
        carry = d.quot;
    }
    while (carry != 0)
    {
        std::div_t d = std::div(carry, 10);
        lhs.push_back(elem_type(d.rem));
        carry = d.quot;
    }
}

//
// Multiplication generates as many digits as it needs.
//
void decimal::mul(absval& lhs, const absval& rhs)
{
    std::vector<absval> parts;

    absval temp;

    for (citer it = rhs.cbegin(), end = rhs.cend(); it != end; ++it)
    {
        temp = lhs;
        mul_part(temp, *it);
        parts.push_back(temp);
    }

    lhs.clear();
    absval::size_type shift = 0;
    for (absval& part : parts)
    {
        part.insert(part.cbegin(), shift, elem_type(0));
        add(lhs, part);
        ++shift;
    }
}

decimal& decimal::operator*=(const decimal& rhs)
{
    if (this->iszero())
    {
        return *this;
    }
    if (rhs.iszero())
    {
        return this->set_to_zero();
    }
    if (&rhs == this) // 2024-01-25:  guard against aliasing
    {
        decimal temp(*this);
        temp *= rhs;
        *this = move(temp);
        return *this;
    }
    //
    // For *= and /=, we'll also check for multiplication and division by one.
    //
    if (rhs.isone())
    {
        neg = this->neg != rhs.neg;
        return *this;
    }
    if (this->isone())
    {
        val = rhs.val;
        neg = this->neg != rhs.neg;
        this_scale = rhs.this_scale;
        return *this;
    }

    //
    // If none of the above, we're stuck with summing partial products.
    //
    mul(this->val, rhs.val);
    neg = this->neg != rhs.neg;
    this_scale += rhs.this_scale;
    trim_trailing_zeros();
    return *this;
}

//
// Find the next digit of the trial quotient by doing a binary search.
// Return the next digit through the call and the new remainder
// through the first argument.
//
decimal::elem_type decimal::div_part(absval& lhs, const absval& rhs,
                                     elem_type min, elem_type max)
{
    if (max > min)
    {
        elem_type trial_quot = (max + min) / 2;
        absval prod_part(rhs);
        mul_part(prod_part, trial_quot);

        // If dividend < quot * divisor, quot needs to be smaller.
        if (raw_cmp(lhs, prod_part) < 0)
        {
            return div_part(lhs, rhs, min, trial_quot);
        }

        absval trial_rem = lhs;
        sub(trial_rem, prod_part);

        // If divisor < rem, quot needs to be bigger.
        if (raw_cmp(rhs, trial_rem) < 0)
        {
            return div_part(lhs, rhs, trial_quot + 1, max);
        }

        // Got a quotient and a remainder.
        lhs = move(trial_rem);
        return trial_quot;
    }
    return 0;
}

//
// Following is the long division algorithm that the author learned
// in the fourth grade at school.
//
// This function returns an int equal to the number of additional LSDs
// beyond lhs.size() to help the caller to determine the scale.
//
int decimal::div(absval& lhs, const absval& rhs)
{
    absval quot;        // what we're computing
    absval quot_part;   // sufficient lhs digits to divide by rhs
    elem_type next_dig; // the next quot digit
    int frac_digs = 0;  // count of digits after lhs' LSD
    int tot_digs = 0;   // quot.size();

    size_t lhsidx = lhs.size();

    //
    // Initialize quot_part with enough MSDs of lhs
    // so that it's >= rhs.  (Remember that absval
    // is a little-endian deque<elem_type>.)
    //
    while (raw_cmp(quot_part, rhs) < 0)
    {
        if (lhsidx > 0)
        {
            quot_part.push_front(lhs[--lhsidx]);
        }
        else
        {
            quot_part.push_front(0);
            ++frac_digs;
        }
    }

    for (;;)
    {
        next_dig = div_part(quot_part, rhs, 0, 10);

        quot.push_front(next_dig);
        ++tot_digs;

        //
        // If we've already generated two guard digits, we're done.
        //
        if (tot_digs > max_digs + 1)
        {
            break;
        }

        //
        // The remainder after partial division is now in quot_part.
        // If we've handled all the digits of lhs and quot_part == 0,
        // we're done.
        //
        while (!quot_part.empty() && quot_part.back() == 0)
        {
            quot_part.pop_back();
        }
        if (lhsidx == 0 && quot_part.empty())
        {
            break;
        }

        //
        // Get the next lhs digit and loop.
        //
        if (lhsidx > 0)
        {
            quot_part.push_front(lhs[--lhsidx]);
        }
        else
        {
            quot_part.push_front(0);
            ++frac_digs;
        }
    } // for(;;)

    lhs = move(quot);

    return frac_digs;
}

decimal& decimal::operator/=(const decimal& rhs)
{
    if (rhs.iszero())
    {
        throw std::domain_error("decimal division by zero");
    }
    if (this->iszero())
    {
        return *this;
    }
    if (&rhs == this) // 2024-01-25:  guard against aliasing
    {
        return this->set_to_one();
    }
    if (rhs.isone())
    {
        neg = this->neg != rhs.neg;
        return *this;
    }
    this_scale += div(val, rhs.val) - rhs.this_scale;
    neg = this->neg != rhs.neg;
    trim_all_zeros();
    do_round(max_digs, rnd);
    return *this;
}

decimal rint(const decimal& val)
{
    decimal temp(val);
    temp.set_user_scale(0);
    // just user-requested rounding
    temp.round();
    return temp;
}

decimal modf(const decimal& val, decimal* iptr)
{
    decimal temp(move(trunc(val)));
    if (iptr != nullptr)
    {
        *iptr = temp;
    }
    return move(val - temp);
}

decimal remainder(const decimal& dividend, const decimal& divisor, rounding rnd)
{
    decimal temp(dividend);
    temp /= divisor;
    temp.set_rounding(rnd);
    temp.round();
    temp.set_user_scale(decimal::floating_scale);
    temp *= divisor;
    // temp = dividend - temp:
    temp.negate();    
    temp += dividend;
    return temp;
}

decimal copysign(const decimal& val, const decimal& sgn)
{
    decimal temp(val);
    if (!val.iszero() && val.isneg() != sgn.isneg())
    {
        temp.negate();
    }
    return temp;
}

} // namespace bignum

// End of decimal.cpp
