//
// rational.cpp
//
// Bill Seymour, 2024-02-05
//
// Copyright Bill Seymour 2024.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This file contains definitions of non-inline functions
// for the rational number library in rational.hpp.
//

#include "rational.hpp"

#include <stdexcept>
#include <cfloat>
#include <cmath>

namespace {

using std::move;

//
// RAII:  set a long double to its absolute value; then on destruction,
// adjust the sign of a rational object to be the original sign of the
// second ctor argument.
//
class sign_setter
{
    bignum::rational* pr;
    bool neg;
public:
    sign_setter(bignum::rational* rat, long double& val)
      : pr(rat), neg(val < 0.0L)
    {
        if (neg)
        {
            val = -val;
        }
    }
    ~sign_setter() noexcept
    {
        if (neg)
        {
            pr->negate();
        }
    }
    sign_setter(const sign_setter&) = delete;
    sign_setter(sign_setter&&) = delete;
    sign_setter& operator=(const sign_setter&) = delete;
    sign_setter& operator=(sign_setter&&) = delete;
};

} // anonymous namespace

namespace bignum {

//
// Private members:
//

const rational::foo rational::bar;

//
// Conversion from C-style strings:
//
void rational::init(const char* s, int radix, const char** endptr)
{
    if (s == nullptr)
    {
        throw std::invalid_argument("Conversion to rational from nullptr");
    }

    bool isneg = false;
    if (*s == '-')
    {
        isneg = true;
        ++s;
    }
    else if (*s == '+')
    {
        ++s;
    }

    if (radix != 8 && radix != 10 && radix != 16)
    {
        if (*s == '0')
        {
            ++s;
            if (std::tolower(*s) == 'x')
            {
                radix = 16;
                ++s;
            }
            else
            {
                radix = 8;
            }
        }
        else
        {
            radix = 10;
        }
    }

    detail::parse_return retval = detail::parse(s, radix);

    if (retval.error)
    {
        throw std::invalid_argument(
            "Parse error converting string to rational");
    }
    if (retval.got_den && retval.den.is_zero())
    {
        throw std::invalid_argument(
            "Converting string with zero denominator to rational");
    }

    //
    // Success:
    //
    num = move(retval.num);
    if (isneg)
    {
        num.negate();
    }

    if (retval.got_den) // we got a '/'
    {
        den = move(retval.den);
    }
    else // no '/'
    {
        den.set_to_one();

        // but there might have been a '.'
        int exp = retval.frac_digs + retval.exponent;
        if (exp > 0)
        {
            num *= std::pow(10, exp);
        }
        else if (exp < 0)
        {
            den *= std::pow(10, -exp);
        }
    }

    if (endptr != nullptr)
    {
        *endptr = retval.term;
    }
}

//
// Conversion from std::strings:
//
void rational::init(const std::string& val, int radix, std::size_t* term)
{
    const char* beg = val.c_str();
    const char* end = nullptr;
    init(beg, radix, &end);
    if (term != nullptr)
    {
        *term = static_cast<std::size_t>(end - beg);
    }
}

//
// Exact conversion from floating point types
// (precondition val >= 0, statically proveable):
//
void rational::init_exact(long double val)
{
    int exp;
    long double mant = std::frexp(val, &exp);

    while (mant != std::floor(mant))
    {
        mant *= FLT_RADIX;
        --exp;
    }
    // mant is now an integer value,
    // and exp has been adjusted accordingly.

    num = mant;
    den.set_to_one();

    if (exp > 0)
    {
      #if FLT_RADIX == 2
        // almost everybody these days
        num <<= exp;
      #elif FLT_RADIX == 16
        // IBM mainframes, I've heard
        num <<= 4 * exp;
      #else
        // probably never, but just in case...
        num *= integer(std::pow(static_cast<long double>(FLT_RADIX), exp));
      #endif
        // den == 1, so already lowest terms
    }
    else if (exp < 0)
    {
      #if FLT_RADIX == 2
        den <<= -exp;
        // den is a power of 2, so...
        if (num.is_even())
        {
            normalize();
        }
      #elif FLT_RADIX == 16
        den <<= 4 * -exp;
        normalize();
      #else
        den =
          move(integer(std::pow(static_cast<long double>(FLT_RADIX), -exp)));
        normalize();
      #endif
    }
}

//
// Conversion from floating point types:
//
void rational::init(long double val, long double err)
{
    if (!std::isfinite(val))
    {
        throw std::invalid_argument(
            "Can't convert non-finite floating point to rational");
    }
    if (val == 0.0L)
    {
        set_to_zero();
        return;
    }

    sign_setter RAII(this, val);

    if (!std::isfinite(err))
    {
        init_exact(val);
        return;
    }
    // else continued fractions (Knuth, 4.5.3):

    long double max_error = std::fabs(val * err);

    long double val0 = val; // current fraction term

    integer int0 = integer(val0);

    integer num0 = int0,  // current approximation
            den0 = 1U;    // first guess:  int(val) / 1

    integer num1 = 1U, num2; // previous and 2nd-previous numerator
    integer den1 = 0U, den2; // previous and 2nd-previous denominator

    long double err0 = std::fabs(val0 - static_cast<long double>(int0));

    while (err0 > max_error)
    {
        val0 = 1.0 / (val0 - static_cast<long double>(int0));
        if (std::isinf(val0))
        {
            break; // the best we can do
        }
        int0 = integer(val0);

        num2 = num1;
        num1 = num0;
        num0 = int0 * num1 + num2;

        den2 = den1;
        den1 = den0;
        den0 = int0 * den1 + den2;

        long double err1 = err0;
        err0 = std::fabs(
         val - static_cast<long double>(num0) / static_cast<long double>(den0));

        if (err0 >= err1) // no longer converging
        {
            break; // the best we can do
        }
    }

    num = move(num0);
    den = move(den0);
    isnorm = true; // continued fractions guarantee that
}

[[noreturn]] void rational::div_by_zero(const char* where)
{
    std::string why("Division by zero in ");
    why.append(where);
    throw std::domain_error(why);
}

//
// Public members:
//

std::string rational::to_string(int radix) const
{
    if (!isnorm)
    {
        normalize();
    }
    std::string retval(move(num.to_string(radix)));
    retval.append(1, '/');
    retval.append(move(den.to_string(radix)));
    return retval;
}

int rational::compare(const rational& rhs) const
{
    std::size_t lsize = this->numer().size() + rhs.denom().size();
    std::size_t rsize = rhs.numer().size() + this->denom().size();
    if (lsize < rsize)
    {
        return -1;
    }
    if (rsize < lsize)
    {
        return +1;
    }
    return (this->numer() * rhs.denom()).compare(rhs.numer() * this->denom());
}

const rational& rational::normalize() const
{
    //
    // Tricky code:  we call this a const member function
    // because the numerical value of the object doesn't
    // change.  We return *this by const reference so that
    // users can't play any other tricks later.
    //
    integer gcdval(move(gcd(num, den)));
    if (!gcdval.is_one())
    {
        const_cast<integer&>(num) /= gcdval;
        const_cast<integer&>(den) /= gcdval;
    }
    const_cast<bool&>(isnorm) = true;
    return *this;
}

rational& rational::operator+=(const rational& rhs)
{
    if (!rhs.is_zero())
    {
        if (this->is_zero())
        {
            *this = rhs;
        }
        else if (&rhs == this) // 2024-01-25:  efficient if aliased
        {
            num += num; // integer guards against aliasing
            isnorm = false;
        }
        else
        {
            num = this->num * rhs.den + rhs.num * this->den;
            if (num.is_zero())
            {
                den.set_to_one();
                isnorm = true;
            }
            else
            {
                this->den *= rhs.den;
                isnorm = false;
            }
        }
    }
    return *this;
}

rational& rational::operator-=(const rational& rhs)
{
    if (!rhs.is_zero())
    {
        if (this->is_zero())
        {
            *this = rhs;
            num.negate();
        }
        else if (&rhs == this) // 2024-01-25:  efficient if aliased
        {
            this->set_to_zero();
        }
        else
        {
            num = this->num * rhs.den - rhs.num * this->den;
            if (num.is_zero())
            {
                den.set_to_one();
                isnorm = true;
            }
            else
            {
                this->den *= rhs.den;
                isnorm = false;
            }
        }
    }
    return *this;
}

rational& rational::operator*=(const rational& rhs)
{
    if (!this->is_zero())
    {
        if (rhs.is_zero())
        {
            num.clear();
            den.set_to_one();
            isnorm = true;
        }
        else // integer guards against aliasing
        {
            this->num *= rhs.num;
            this->den *= rhs.den;
            isnorm = false;
        }
    }
    return *this;
}

rational& rational::operator/=(const rational& rhs)
{
    if (!rhs.is_zero())
    {
        if (&rhs == this) // 2024-01-25:  efficient if aliased
        {
            this->set_to_one();
        }
        else if (!this->is_zero())
        {
            this->num *= rhs.den;
            this->den *= rhs.num;
            fix_sign();
            isnorm = false;
        }
        return *this;
    }
    div_by_zero("operator/=(rational)");
}

//
// Non-members:
//

rational round(const rational& val, rounding rnd)
{
    rational temp(val);
    if (!temp.is_norm())
    {
        temp.normalize();
    }

    if (temp.denom().is_one())
    {
        return temp; // it's already an integer
    }

    bool neg = temp.is_neg();

    integer quot(temp.numer());
    quot /= temp.denom(); // truncates toward zero

    //
    // Whether the result is <, ==, or > n/2 (not needed for all_to...):
    //
    bool eq = false;
    bool lt = false;
    bool gt = false;
    if (rnd >= tie_to_neg_inf)
    {
        //
        // Pathological coupling:  any tie_to... > any all_to...
        //
        eq = temp.denom().compare(2) == 0;
        lt = !eq && abs(temp.numer()) % temp.denom() < temp.denom() / 2;
        gt = !eq && !lt;
    }

    switch (rnd)
    {
        case all_to_neg_inf:
            if (neg)
            {
                --quot;
            }
            break;
        case all_to_pos_inf:
            if (!neg)
            {
                ++quot;
            }
            break;
        case all_to_zero:
        case all_fastest:
        case all_smallest:
        case all_unspecified:
            break;
        case all_away_zero:
            if (neg)
            {
                --quot;
            }
            else
            {
                ++quot;
            }
            break;
        case all_to_odd:
            if (!quot.is_odd())
            {
                if (neg)
                {
                    --quot;
                }
                else
                {
                    ++quot;
                }
            }
            break;
        case all_to_even:
            if (quot.is_odd())
            {
                if (neg)
                {
                    --quot;
                }
                else
                {
                    ++quot;
                }
            }
            break;

        case tie_to_neg_inf:
            if (neg && !lt)
            {
                --quot;
            }
            else if (!neg & gt)
            {
                ++quot;
            }
            break;
        case tie_to_pos_inf:
            if (!neg && !lt)
            {
                ++quot;
            }
            else if (neg && lt)
            {
                --quot;
            }
            break;
        case tie_away_zero:
            if (neg && !lt)
            {
                --quot;
            }
            else if (!neg && !lt)
            {
                ++quot;
            }
            break;
        case tie_to_odd:
            if (gt || eq && !quot.is_odd())
            {
                if (neg)
                {
                    --quot;
                }
                else
                {
                    ++quot;
                }
            }
            break;
        case tie_to_even:
            if (gt || eq && quot.is_odd())
            {
                if (neg)
                {
                    --quot;
                }
                else
                {
                    ++quot;
                }
            }
            break;
        default: // tie_fastest, tie_smallest, tie_unspecified, tie_to_zero
            if (gt)
            {
                if (neg)
                {
                    --quot;
                }
                else
                {
                    ++quot;
                }
            }
            break;
    }

    temp = move(quot);
    return temp;
}

rational modf(const rational& val, rational* intptr)
{
    rational temp(move(trunc(val)));
    if (intptr != nullptr)
    {
        *intptr = temp;
    }
    return move(val - temp);
}

rational remainder(const rational& dividend,
                   const rational& divisor,
                   rounding rnd)
{
    rational temp(dividend);
    temp /= divisor;
    temp = move(round(temp, rnd));
    temp *= divisor;
    // temp = dividend - temp:
    temp.negate();    
    temp += dividend;
    return temp;
}

rational pow(const rational& val, int exp)
{
    rational result;
    result.set_to_one();

    if (exp != 0)
    {
        rational temp(val);
        unsigned x = std::abs(exp);

        for (;;) // Knuth, 4.6.3
        {
            if ((x & 1U) != 0U)
            {
                result *= temp;
            }

            if ((x >>= 1U) == 0U)
            {
                break;
            }

            temp *= temp;
        }

        if (exp < 0)
        {
            result.invert();
        }
    }

    return result;
}

} // namespace bignum

// End of rational.cpp
