//
// rational.hpp
//
// Bill Seymour, 2024-02-02
//
// Copyright Bill Seymour 2024.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This is the header for an open-source rational number.
//

#ifndef RATIONAL_HPP_INCLUDED
#define RATIONAL_HPP_INCLUDED

#include "integer.hpp" // #includes everything we need

namespace bignum {

namespace detail // no special header for just this
{
    struct parse_return
    {
        integer num, den;
        integer* pval = &num;
        int exponent = 0, frac_digs = 0;
        bool got_den = false, negexp = false, error = false;
        const char* term = nullptr;
    };
    parse_return parse(const char*, int);
}

class rational final
{
    integer num, den;
    bool isnorm;

    //
    // For SFINAE when converting from a single built-in integer:
    //
    struct foo { };
    static const foo bar;

    //
    // For SFINAE when converting from pairs of built-in integers:
    //
    template<typename T, typename U>
    struct both_int : std::integral_constant<bool,
        std::is_integral<T>::value && std::is_integral<U>::value> { };

    //
    // Helpers for conversions from floating point types:
    //
    void init_exact(long double);
    void init(long double, long double);

    //
    // Helpers for conversions from strings:
    //
    void init(const char*, int, const char**);
    void init(const std::string&, int, std::size_t*);

    //
    // There are a few places where we could get a negative denominator:
    //
    void fix_sign() noexcept
    {
        if (den.is_neg())
        {
            num.negate();
            den.negate();
        }
    }

    //
    // Throws a domain_error exception:
    //
    [[noreturn]] static void div_by_zero(const char*);

public:
    //
    // Special member functions and swap():
    //
    rational() : num(), den(), isnorm(true) { den.set_to_one(); }

    ~rational() noexcept { }

    rational(const rational&) = default;
    rational(rational&&) = default;

    rational& operator=(const rational&) = default;
    rational& operator=(rational&&) = default;

    void swap(rational& other)
    {
        num.swap(other.num);
        den.swap(other.den);
        std::swap(isnorm, other.isnorm);
    }

    //
    // Conversion from one or two built-in integers:
    //
    // We can't just use enable_if<...,built-in-type> in the one-arg ctor
    // because of ambiguity with both the two-arg version and conversions
    // from floating point types, so we'll use a private type.
    //
    template<typename I>
    rational(I val,
      typename std::enable_if<std::is_integral<I>::value,foo>::type = bar)
      : num(val), den(), isnorm(true)
    {
        den.set_to_one();
    }
    template<typename N, typename D>
    rational(N nval, D dval,
      typename std::enable_if<both_int<N,D>::value,int>::type = 0)
      : num(nval), den(dval)
    {
        if (dval == 0)
        {
            div_by_zero("rational(intT,intT)");
        }
        fix_sign();
        normalize();
    }

    //
    // We'll also allow conversions from integers:
    //
    rational(const integer& val) : num(val), den(), isnorm(true)
    {
        den.set_to_one();
    }
    rational(integer&& val) : num(std::move(val)), den(), isnorm(true)
    {
        den.set_to_one();
    }
    rational(const integer& nval, const integer& dval) : num(nval), den(dval)
    {
        if (dval.is_zero())
        {
            div_by_zero("rational(integer,integer)");
        }
        fix_sign();
        normalize();
    }
    rational(integer&& nval, integer&& dval)
      : num(std::move(nval)), den(std::move(dval))
    {
        if (den.is_zero()) // NB: den, not dval, because we've already moved it
        {
            div_by_zero("rational(integer,integer)");
        }
        fix_sign();
        normalize();
    }

    //
    // Explicit conversion from floating point types:
    //
    template<typename F>
    explicit rational(F val, long double err = 0.0,
      typename std::enable_if<std::is_floating_point<F>::value,int>::type = 0)
      : num(), den()
    {
        init(static_cast<long double>(val), err);
    }

    //
    // Explicit conversion from strings:
    //
    explicit rational(const char* val,
                      int radix = 0,
                      const char** end = nullptr)
      : num(), den()
    {
        init(val, radix, end);
    }
    explicit rational(const std::string& val,
                      int radix = 0,
                      std::size_t* term = nullptr)
      : num(), den()
    {
        init(val, radix, term);
    }

    //
    // Assignment:
    //
    template<typename I>
    typename std::enable_if<std::is_integral<I>::value,rational>::type&
    operator=(I val)
    {
        num = val;
        den.set_to_one();
        isnorm = true;
        return *this;
    }
    template<typename N, typename D>
    typename std::enable_if<both_int<N,D>::value,rational>::type&
    assign(N nval, D dval)
    {
        if (dval == 0)
        {
            div_by_zero("assign(intT,intT)");
        }
        num = nval;
        den = dval;
        fix_sign();
        normalize();
        return *this;
    }
    rational& operator=(const integer& val)
    {
        num = val;
        den.set_to_one();
        isnorm = true;
        return *this;
    }
    rational& operator=(integer&& val)
    {
        num = std::move(val);
        den.set_to_one();
        isnorm = true;
        return *this;
    }
    rational& assign(const integer& nval, const integer& dval)
    {
        if (dval.is_zero())
        {
            div_by_zero("assign(integer,integer)");
        }
        num = nval;
        den = dval;
        fix_sign();
        normalize();
        return *this;
    }
    rational& assign(integer&& nval, integer&& dval)
    {
        if (dval.is_zero())
        {
            div_by_zero("assign(integer,integer)");
        }
        num = std::move(nval);
        den = std::move(dval);
        fix_sign();
        normalize();
        return *this;
    }
    template<typename F>
    typename std::enable_if<std::is_floating_point<F>::value,rational>::type&
    assign(F val, long double err = 0.0)
    {
        init(static_cast<long double>(val), err);
        return *this;
    }
    rational& assign(const char* val,
                     int radix = 0,
                     const char** end = nullptr)
    {
        init(val, radix, end);
        return *this;
    }
    rational& assign(const std::string& val,
                     int radix = 0,
                     std::size_t* term = nullptr)
    {
        init(val, radix, term);
        return *this;
    }

    //
    // Observers:
    //
    const integer& numer() const noexcept
    {
        return num;
    }
    const integer& denom() const noexcept
    {
        return den;
    }

    bool is_zero() const noexcept
    {
        return num.is_zero();
    }
    bool is_neg() const noexcept
    {
        return num.is_neg();
    }
    bool is_pos() const noexcept
    {
        return !is_neg() && !is_zero();
    }
    bool is_one() const noexcept
    {
        return bignum::abs(num) == den;
    }
    bool is_one(bool n) const noexcept
    {
        return is_neg() == n && is_one();
    }
    bool is_norm() const noexcept
    {
        return isnorm;
    }

    explicit operator bool() const noexcept
    {
        return !num.is_zero();
    }

    explicit operator long double() const
    {
        if (!isnorm)
        {
            normalize();
        }
        return static_cast<long double>(num) / static_cast<long double>(den);
    }

    std::string to_string(int = 10) const;

    int compare(const rational&) const;

    int signum() const noexcept
    {
        return is_neg() ? -1 : is_zero() ? 0 : +1;
    }

    //
    // Mutators:
    //
    rational& set_to_zero()
    {
        num.clear();
        den.set_to_one();
        isnorm = true;
        return *this;
    }

    rational& set_to_one(bool neg = false)
    {
        num.set_to_one(neg);
        den.set_to_one();
        isnorm = true;
        return *this;
    }

    rational& negate() noexcept
    {
        if (!is_zero())
        {
            num.negate();
        }
        return *this;
    }

    rational& abs() noexcept
    {
        num.abs();
        return *this;
    }

    rational& invert()
    {
        if (!num.is_zero())
        {
            num.swap(den);
            fix_sign();
            return *this;
        }
        div_by_zero("invert()");
    }

    const rational& normalize() const;

    //
    // Arithmetic operators:
    //
    rational& operator++()
    {
        num += den;
        return *this;
    }
    rational& operator--()
    {
        num -= den;
        return *this;
    }

    rational operator++(int)
    {
        rational temp(*this);
        num += den;
        return temp;
    }
    rational operator--(int)
    {
        rational temp(*this);
        num -= den;
        return temp;
    }

    rational& operator+=(const rational&);
    rational& operator-=(const rational&);
    rational& operator*=(const rational&);
    rational& operator/=(const rational&);

    //
    // Heap usage:
    //
    std::size_t size() const noexcept
    {
        return num.size() + den.size();
    }
    std::size_t capacity() const noexcept
    {
        return num.capacity() + den.capacity();
    }
    void shrink_to_fit()
    {
        num.shrink_to_fit();
        den.shrink_to_fit();
    }
};

//
// Non-member functions:
//
using std::swap;
inline void swap(rational& lhs, rational& rhs)
{
    lhs.swap(rhs);
}

inline std::string to_string(const rational& val, int radix = 10)
{
    return std::move(val.to_string(radix));
}

inline rational reciprocal(const rational& val)
{
    rational temp(val);
    temp.invert();
    return temp;
}

//
// We don't have NaNs or infinities, so rationals are strongly ordered.
//
inline bool operator==(const rational& lhs, const rational& rhs)
{
    return lhs.compare(rhs) == 0;
}
inline bool operator!=(const rational& lhs, const rational& rhs)
{
    return lhs.compare(rhs) != 0;
}
inline bool operator<(const rational& lhs, const rational& rhs)
{
    return lhs.compare(rhs) < 0;
}
inline bool operator>(const rational& lhs, const rational& rhs)
{
    return lhs.compare(rhs) > 0;
}
inline bool operator<=(const rational& lhs, const rational& rhs)
{
    return lhs.compare(rhs) <= 0;
}
inline bool operator>=(const rational& lhs, const rational& rhs)
{
    return lhs.compare(rhs) >= 0;
}

//
// Non-member operators:
//
inline rational operator+(const rational& rhs)
{
    return rhs;
}
inline rational operator-(const rational& rhs)
{
    rational temp(rhs);
    temp.negate();
    return temp;
}
inline rational operator+(const rational& lhs, const rational& rhs)
{
    rational temp(lhs);
    temp += rhs;
    return temp;
}
inline rational operator-(const rational& lhs, const rational& rhs)
{
    rational temp(lhs);
    temp -= rhs;
    return temp;
}
inline rational operator*(const rational& lhs, const rational& rhs)
{
    rational temp(lhs);
    temp *= rhs;
    return temp;
}
inline rational operator/(const rational& lhs, const rational& rhs)
{
    rational temp(lhs);
    temp /= rhs;
    return temp;
}

//
// A few <cmath>-like functions:
//
inline rational abs(const rational& val)
{
    rational temp(val);
    temp.abs();
    return temp;
}

//
// We'll pass a rounding mode to bignum::round() and then use that to implement
// several other functions.  The default, tie_away_zero, mimics std::round().
//
rational round(const rational&, rounding = tie_away_zero);

inline rational ceil(const rational& val)
{
    return std::move(round(val, all_to_pos_inf));
}

inline rational floor(const rational& val)
{
    return std::move(round(val, all_to_neg_inf));
}

inline rational trunc(const rational& val)
{
    return std::move(round(val, all_to_zero));
}

inline rational rint(const rational& val)
{
    return std::move(round(val, tie_to_even));
}

rational modf(const rational&, rational* = nullptr);

//
// We'll also allow passing a rounding mode to remainder();
// tie_to_even mimics std::remainder().
//
rational remainder(const rational&, const rational&, rounding = tie_to_even);

inline rational fmod(const rational& x, const rational& y)
{
    return std::move(remainder(x, y, all_to_zero));
}

inline rational sqr(const rational& val)
{
    rational temp(val);
    temp *= val;
    return temp;
}

rational pow(const rational&, int);

inline rational copysign(const rational& lhs, const rational& rhs)
{
    rational temp(lhs);
    if (lhs.is_neg() != rhs.is_neg())
    {
        temp.negate();
    }
    return temp;
}

inline rational fma(const rational& multiplicand,
                    const rational& multiplier,
                    const rational& addend)
{
    rational temp(multiplicand);
    temp *= multiplier;
    temp += addend;
    return temp;
}

} // namespace bignum

#ifndef RATIONAL_NO_IO_OPERATORS
  #include "rational_io.hpp"
#endif

#endif // RATIONAL_HPP_INCLUDED
