//
// rational_io.hpp
//
// Bill Seymour, 2024-02-02
//
// Copyright Bill Seymour 2024.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This header provides the definitions of the << and >> I/O operators
// for the rational class defined in rational.hpp, in which file this
// header has already been #included if necessary.
//

#ifndef RATIONAL_IO_HPP_INCLUDED
#define RATIONAL_IO_HPP_INCLUDED

#include <iostream>
#include <sstream>

namespace bignum {

namespace detail {

//
// For access to ios_base's iword and badbit:
//
struct ioswrap
{
    virtual long& iword(int) const = 0;
    virtual bool bad() const = 0;
};
template<typename Ch, typename Tr>
class iosleaf : public ioswrap
{
    std::basic_ios<Ch,Tr>* iosp;
public:
    explicit iosleaf(std::basic_ios<Ch,Tr>& ios) : iosp(&ios) { }
    long& iword(int i) const { return iosp->iword(i); }
    bool bad() const { return iosp->bad(); }
};

//
// Indexes for ios_base::iword...not thread-safe initialization,
// but iostreams aren't thread-safe anyway:
//
const int genidx = std::ios_base::xalloc(); // generic for no-arg manips
const int dividx = std::ios_base::xalloc(); // for setdiv

//
// Set, clear and test iword bits:
//
inline void setbit(const ioswrap& ios, long bit)
{
    long& iwd = ios.iword(genidx);
    if (!ios.bad())
    {
        iwd |= bit;
    }
}
inline void clearbit(const ioswrap& ios, long bit)
{
    long& iwd = ios.iword(genidx);
    if (!ios.bad())
    {
        iwd &= ~bit;
    }
}
template<typename Ch, typename Tr>
inline bool is_set(std::basic_ostream<Ch,Tr>& os, long bit)
{
    long& iwd = os.iword(genidx);
    return !os.bad() && (iwd & bit) != 0L;
}

constexpr long denbit = 1L;
constexpr long adjbit = 2L;

//
// showden1 helpers:
//
inline void showden1(const ioswrap& ios)
{
    setbit(ios, denbit);
}
inline void noshowden1(const ioswrap& ios)
{
    clearbit(ios, denbit);
}
template<typename Ch, typename Tr>
inline bool showden1_required(std::basic_ostream<Ch,Tr>& os)
{
    return is_set(os, denbit);
}

//
// divalign helpers:
//
inline void divalign(const ioswrap& ios)
{
    setbit(ios, adjbit);
}
inline void nodivalign(const ioswrap& ios)
{
    clearbit(ios, adjbit);
}
template<typename Ch, typename Tr>
inline bool divalign_required(std::basic_ostream<Ch,Tr>& os)
{
    return is_set(os, adjbit);
}

//
// Get the division sign:
//
template<typename Ch, typename Tr>
Ch divsign(std::basic_ios<Ch,Tr>& ios)
{
    long& iwd = ios.iword(dividx);
    if (!ios.bad())
    {
        if (iwd == 0L)
        {
            iwd = ios.widen('/');
        }
        return Ch(iwd);
    }
    return ios.widen('/');
}

//
// The return type for setdiv(Ch) and its overloaded >> and <<:
//
class divsetter
{
    long val;
public:
    explicit divsetter(long c) : val(c) { }
    void operator()(const ioswrap& ios) const
    {
        long& iwd = ios.iword(dividx);
        if (!ios.bad())
        {
            iwd = val;
        }
    }
};
template<typename Ch, typename Tr>
std::basic_istream<Ch,Tr>&
operator>>(std::basic_istream<Ch,Tr>& is, const divsetter& ds)
{
    ds(iosleaf<Ch,Tr>(is));
    return is;
}
template<typename Ch, typename Tr>
std::basic_ostream<Ch,Tr>&
operator<<(std::basic_ostream<Ch,Tr>& os, const divsetter& ds)
{
    ds(iosleaf<Ch,Tr>(os));
    return os;
}

//
// RAII for clearing and resetting a flag:
//
class ioflag_resetter
{
    std::ios_base* p;
    std::ios_base::fmtflags f;
public:
    ioflag_resetter(std::ios_base& iob, std::ios_base::fmtflags flg)
      : p(&iob), f(iob.flags())
    {
        iob.unsetf(flg);
    }
    ~ioflag_resetter()
    {
        p->flags(f);
    }
    ioflag_resetter(const ioflag_resetter&) = delete;
    ioflag_resetter(ioflag_resetter&&) = delete;
    ioflag_resetter& operator=(const ioflag_resetter&) = delete;
    ioflag_resetter& operator=(ioflag_resetter&&) = delete;
};

} // namespace detail

//
// The manipulators:
//
template<typename Ch, typename Tr>
std::basic_ostream<Ch,Tr>& showden1(std::basic_ostream<Ch,Tr>& os)
{
    detail::showden1(detail::iosleaf<Ch,Tr>(os));
    return os;
}
template<typename Ch, typename Tr>
std::basic_ostream<Ch,Tr>& noshowden1(std::basic_ostream<Ch,Tr>& os)
{
    detail::noshowden1(detail::iosleaf<Ch,Tr>(os));
    return os;
}
template<typename Ch, typename Tr>
std::basic_ostream<Ch,Tr>& divalign(std::basic_ostream<Ch,Tr>& os)
{
    detail::divalign(detail::iosleaf<Ch,Tr>(os));
    return os;
}
template<typename Ch, typename Tr>
std::basic_ostream<Ch,Tr>& nodivalign(std::basic_ostream<Ch,Tr>& os)
{
    detail::nodivalign(detail::iosleaf<Ch,Tr>(os));
    return os;
}
template<typename Ch>
const detail::divsetter setdiv(Ch c)
{
    return detail::divsetter(static_cast<long>(c));
}

//
// The operators:
//

template<typename Ch, typename Tr>
std::basic_istream<Ch,Tr>&
operator>>(std::basic_istream<Ch,Tr>& istr, rational& rat)
{
    if (!istr.good())
    {
        return istr;
    }

    integer num;
    if (!(istr >> num).good())
    {
        return istr;
    }
    if (istr.peek() != detail::divsign(istr))
    {
        rat = std::move(num);
        return istr;
    }
    (void)istr.get(); // remove '/'

    detail::ioflag_resetter iofr(istr, std::ios_base::skipws);

    integer den;
    if (!(istr >> den).good())
    {
        return istr;
    }
    if (den.is_zero())
    {
        istr.setstate(std::ios_base::failbit);
        return istr;
    }

    rat.assign(std::move(num), std::move(den));
    return istr;
}

template<typename Ch, typename Tr>
std::basic_ostream<Ch,Tr>&
operator<<(std::basic_ostream<Ch,Tr>& ostr, const rational& rat)
{
    if (!rat.is_norm())
    {
        rat.normalize();
    }

    if (rat.denom().is_one() && !detail::showden1_required(ostr))
    {
        ostr << rat.numer();
    }
    else
    {
        typename std::basic_ostream<Ch,Tr>::sentry sen(ostr);
        detail::ioflag_resetter iofr(ostr, std::ios_base::unitbuf);

        if (detail::divalign_required(ostr))  // ostr.width() applies to numerator only
        {
            ostr << rat.numer();
            ostr.put(detail::divsign(ostr));
            ostr.unsetf(std::ios_base::showpos | std::ios_base::showbase);
            ostr << rat.denom();
        }
        else  // write to stringstream first, then align whole fraction
        {
            std::basic_ostringstream<Ch,Tr> temp;
            temp.imbue(ostr.getloc());
            temp.copyfmt(ostr);
            temp.width(0);
            temp.unsetf(std::ios_base::showpos | std::ios_base::showbase);

            const std::ios_base::fmtflags flags = ostr.flags();
            const std::ios_base::fmtflags base  = flags & std::ios_base::basefield;

            integer num = rat.numer();
            char sign = '\0';
            if (base == std::ios_base::dec)
            {
                if (num.is_neg())
                {
                    num.negate();
                    sign = '-';
                }
                else if ((flags & std::ios_base::showpos) != 0)
                {
                    sign = '+';
                }
            }

            temp << num;
            temp.put(detail::divsign(ostr));
            temp << rat.denom();

            int fill_width = int(ostr.width(0)) - int(temp.str().size());
            const Ch fill_char = ostr.fill();

            const bool right_align = (flags & std::ios_base::right) != 0;
            const bool left_align =
                (flags & std::ios_base::adjustfield) == std::ios_base::left;

            if (sign != '\0')
            {
                --fill_width;
                if (right_align)
                    for ( ; fill_width > 0; --fill_width)
                        ostr.put(fill_char);
                ostr.put(ostr.widen(sign));
            }
            else if ((flags & std::ios_base::showbase) != 0)
            {
                if (base == std::ios_base::oct)
                {
                    --fill_width;
                    if (!left_align) // right or internal
                        for ( ; fill_width > 0; --fill_width)
                            ostr.put(fill_char);
                    ostr.put(ostr.widen('0'));
                }
                else if (base == std::ios_base::hex)
                {
                    fill_width -= 2;
                    if (right_align) // but not internal
                        for ( ; fill_width > 0; --fill_width)
                            ostr.put(fill_char);
                    ostr.put(ostr.widen('0'));
                    ostr.put(ostr.widen((flags & std::ios_base::uppercase) != 0 ?
                                        'X' : 'x'));
                }
            }

            if (!left_align)
                for ( ; fill_width > 0; --fill_width)
                    ostr.put(fill_char);

            ostr << temp.str();

            for ( ; fill_width > 0; --fill_width)
                ostr.put(fill_char);
        }
    }

    return ostr;
}

} // namespace bignum

#endif // RATIONAL_IO_HPP_INCLUDED
