//
// rational_parse.cpp
//
// Bill Seymour, 2024-01-03
//
// Copyright Bill Seymour 2024.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//
// This is the implementation of the parser that's used for
// construction and assignment of strings to rationals.
// It's in its own source file just to separate it from
// the more central code for doing rational arithmetic.
//

#include "rational.hpp"

#include <cctype>

namespace {

using bignum::detail::parse_return;

enum class token_type : unsigned char {dig, div, pnt, ths, exp, sgn, oth, cnt};

token_type lex(unsigned char c, int radix)
{
    if (std::isdigit(c) && (radix != 8 || c < '8') ||
        radix == 16 && std::isxdigit(c))
    {
        return token_type::dig;
    }

    static constexpr token_type toktyp[] =
    {
      /*00*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*04*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*08*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*0C*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*10*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*14*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*18*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*1C*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*20*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*24*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
                                                                   // '+'
      /*28*/ token_type::oth, token_type::oth, token_type::oth, token_type::sgn,
                // ','              '-'              '.'              '/'
      /*2C*/ token_type::ths, token_type::sgn, token_type::pnt, token_type::div,
      /*30*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*34*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*38*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*3C*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*40*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
                                 // 'E'
      /*44*/ token_type::oth, token_type::exp, token_type::oth, token_type::oth,
      /*48*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*4C*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*50*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*54*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*58*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*5C*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
      /*60*/ token_type::oth, token_type::oth, token_type::oth, token_type::oth,
                                 // 'e'
      /*64*/ token_type::oth, token_type::exp,
    };
    static constexpr unsigned char maxtok = sizeof toktyp / sizeof toktyp[0];

    return c >= maxtok ? token_type::oth : toktyp[c];
}

//
// Actions that we can take:
//
// Add a digit to the numerator or the denominator.
//
void add(unsigned char c, int radix, parse_return& retval)
{
    static constexpr int charval[] =
    {
      /*00*/ 0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
      /*10*/ 0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
      /*20*/ 0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
      /*30*/ 0, 1, 2, 3,  4, 5, 6, 7,  8, 9, 0, 0,  0, 0, 0, 0, // '0'-'9'

      /*40*/ 0,10,11,12, 13,14,15, 0,  0, 0, 0, 0,  0, 0, 0, 0, // 'A'-'F'
      /*50*/ 0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,  0, 0, 0, 0,
      /*60*/ 0,10,11,12, 13,14,15,                              // 'a'-'f'
    };

    *retval.pval *= radix;
    *retval.pval += charval[c];
}
//
// After a radix point, count a fractional digit as well.
//
void cad(unsigned char c, int radix, parse_return& retval)
{
    ++retval.frac_digs;
    add(c, radix, retval);
}
//
// After 'E' or 'e', add an exponent digit (always decimal).
//
void ead(unsigned char c, int, parse_return& retval)
{
    retval.exponent *= 10;
    retval.exponent += c - '0';
}
//
// A '-' after 'E' or 'e' negates the exponent.
//
void neg(unsigned char c, int, parse_return& retval)
{
    retval.negexp = c == '-';
}
//
// After a division sign, switch to accumulating a denominator.
//
void den(unsigned char, int, parse_return& retval)
{
    retval.pval = &retval.den;
    retval.got_den = true;
}

typedef void (*action)(unsigned char, int, parse_return&);

constexpr action nul = nullptr;

//
// Exit states:
//
constexpr int ok = -1, er = -2;

const struct parse_node
{
    int next_state;
    action act;
}
parse_table[][static_cast<unsigned>(token_type::cnt)] =
{
//state    dig       div       pnt       ths       exp       sgn       oth
  /*0*/{ { 1,add}, {er,nul}, { 4,nul}, { 1,nul}, {er,nul}, {er,nul}, {er,nul} },
  /*1*/{ { 1,add}, { 2,den}, { 5,nul}, { 1,nul}, { 6,nul}, {er,nul}, {ok,nul} },
  /*2*/{ { 3,add}, {er,nul}, {er,nul}, { 2,nul}, {er,nul}, {er,nul}, {er,nul} },
  /*3*/{ { 3,add}, {er,nul}, {er,nul}, { 3,nul}, {er,nul}, {er,nul}, {ok,nul} },
  /*4*/{ { 5,cad}, {er,nul}, {er,nul}, { 4,nul}, {er,nul}, {er,nul}, {er,nul} },
  /*5*/{ { 5,cad}, {er,nul}, {er,nul}, { 5,nul}, { 6,nul}, {er,nul}, {ok,nul} },
  /*6*/{ { 8,ead}, {er,nul}, {er,nul}, { 6,nul}, {er,nul}, { 7,neg}, {er,nul} },
  /*7*/{ { 8,ead}, {er,nul}, {er,nul}, { 7,nul}, {er,nul}, {er,nul}, {er,nul} },
  /*8*/{ { 9,ead}, {er,nul}, {er,nul}, {er,nul}, {er,nul}, {er,nul}, {er,nul} },
  /*9*/{ { 9,ead}, {er,nul}, {er,nul}, {er,nul}, {er,nul}, {er,nul}, {ok,nul} },
};

} // anonymous namespace

namespace bignum {
namespace detail {

parse_return parse(const char* s, int radix)
{
    parse_return retval;

    int state = 0;

    for (;;)
    {
        unsigned char uc = static_cast<unsigned char>(*s);
 
        const parse_node& node =
            parse_table[state][static_cast<unsigned>(lex(uc, radix))];

        if (node.act != nul)
        {
            node.act(uc, radix, retval);
        }
        if ((state = node.next_state) < 0)
        {
            break;
        }
        ++s;
    }

    if (state != ok)
    {
        retval.error = true;
    }
    else if (retval.negexp && retval.exponent > 0)
    {
        retval.exponent = -retval.exponent;
    }

    retval.term = s;

    return retval;
}

} // namespace detail
} // namespace bignum

// End of rational_parse.cpp
